#! /usr/bin/env python

# import standard modules
import os, time

# import custom modules
from utils.error import MyError
from utils.general import TimeSpent, GetCommand
from utils.messages import LogMsg, DebugMsg
from utils.files_paths import FileBoxCls
from utils.subprocesses import RunCommandFromList
from alignment_processing.alignment_functions import (ParseAlignmentFile,
  CalcOverlap, FixAlign)

# constants
MIN_REGION_OVERLAP_FRACTION = 0.70
MAX_MATCH_DIFF = 10

class RealignerCls: #{
  def __init__(self, options, log_info): #{
    self.options = options
    self.log_info = log_info
    # contigs[ctg_id] = realign result
    #   realign result: best alignment of contig to transcripts
    self.contigs = dict()
    self.missing = set()
    #self.contig_seqs = dict()
  #} end def

  #def UpdateContigs(self, group, good_members, store_seq): #{
  def UpdateContigs(self, group, good_members, event_type): #{
    for index in good_members: #{
      ctg_id = group.members[index].contig_info.id
      if (ctg_id not in self.contigs): #{
        self.contigs[ctg_id] = RealignedContigCls(ctg_id)
      #} end if
      self.contigs[ctg_id].UpdateMembers(group.members[index], event_type)
      self.missing.add(ctg_id)
      #if (store_seq): #{
      #  self.contig_seqs[ctg_id] = None
      #} end if
    #} end for
  #} end def

  def RealignContigs(self): #{
    start = time.time()
    LogMsg(self, "Performing contig realignment...")
    # create a query file containing the correct contig sequences
    query_path = self.CreateQueryFile()
    # run the blat alignment without the "-fine" option
    results_path = self.RunBlatAlignment(query_path, use_fine=False)
    # run the blat alignment with the "-fine" option
    fine_results_path = self.RunBlatAlignment(query_path, use_fine=True)
    # parse the alignment results
    self.ParseRealignments(results_path)
    self.ParseRealignments(fine_results_path)
    LogMsg(self, "Time spent performing realignment: %s" %
      TimeSpent(start))
  #} end def

  def CreateQueryFile(self): #{
    LogMsg(self, "Creating query file...")
    query_file_name = "%s.event_contigs.fa" % self.options.lib
    query_path = os.path.join(self.options.realign_dir, query_file_name)
    query_file = FileBoxCls(query_path, "w", "cannot create query contig "
      "sequences file")
    all_contigs_file = FileBoxCls(self.options.ctg_seq_path, "r",
      "cannot read contig sequences file")
    seqs_found = False
    num_written = 0
    for id_line in all_contigs_file: #{
      seq_line = all_contigs_file.next()
      if (not id_line.startswith(">")): #{
        raise RealignerError("invalid contig id line in sequece file:\n%s" %
          id_line)
      #} end if
      # extract the contig id from the line
      ctg_id = id_line.lstrip(">").split()[0]
      #DebugMsg(self, "Contig ID from sequence file: %s" % ctg_id)
      # if the contig is represented in one of the potential predictions
      if (ctg_id in self.contigs): #{
        #DebugMsg(self, "Writing sequence to query file.")
        # write it to the query file
        query_file.WriteLine(id_line)
        query_file.WriteLine(seq_line)
        seqs_found = True
        num_written += 1
        self.contigs[ctg_id].written = True
        if ("itd" in self.contigs[ctg_id].types): #{
          self.contigs[ctg_id].sequence = seq_line
        #} end if
        self.missing.discard(ctg_id)
      #} end if
      #if (ctg_id in self.contig_seqs): #{
      #  self.contig_seqs[ctg_id] = seq_line
      #} end if
    #} for
    if (not seqs_found): #{
      raise RealignerError("could not find any contig sequences in %s" %
        self.options.ctg_seq_path)
    #} end if
    if (num_written != len(self.contigs)): #{
      #missed = list()
      #for contig in self.contigs.itervalues(): #{
      #  if (not contig.written): #{
      #    missed.append(contig.id)
      #  #} end if
      #} end for
      LogMsg(self, "WARNING: only wrote %i of %i contig sequences! " %
        (num_written, len(self.contigs)) + "Missing: %s" %
        ",".join(sorted(self.missing)))
      #  ",".join(missed))
    #} end if
    all_contigs_file.Close()
    query_file.Close()
    return query_path
  #} end def

  def RunBlatAlignment(self, query_path, use_fine=False): #{
    LogMsg(self, "Realigning contigs to transcript sequences...")
    results_file_name = "%s.realign.psl" % self.options.lib
    if (use_fine): #{
      results_file_name = "%s.realign.fine.psl" % self.options.lib
    #} end if
    results_path = os.path.join(self.options.realign_dir, results_file_name)
    # use "stepSize=5" to allow short matches
    # use "repMatch=2253" to mimic webBLAT results
    # use "fine" option to avoid strange misalignments around single mismatches
    command = [GetCommand(self, "blat"), "-stepSize=5", "-repMatch=2253",
      "-minScore=4", self.options.tran_seq_path, query_path, results_path]
    if (use_fine): #{
      command.append("-fine")
    #} end if
    LogMsg(self, " ".join(command))
    return_code = RunCommandFromList(command, dpt=self.options.dpt)
    if (0 != return_code and not self.log_info['debug']): #{
      LogMsg(self, " ".join(command))
    #} end if
    if (0 > return_code): #{
      raise RealignerError("Realignment command was terminated by signal "
        "%i" % return_code)
    elif (0 < return_code):
      raise RealignerError("Error running realignment command: %i" %
        return_code)
    #} end if
    return results_path
  #} end def

  def ParseRealignments(self, results_path): #{
    LogMsg(self, "Parsing realignment results...")
    #filters = { 'count': 50, 'identity': 90.0 }
    filters = { 'identity': 40.0 }
    aligns = ParseAlignmentFile(results_path, filters, self.log_info)
    for align in map(FixAlign, aligns): #{
      ctg_id = align.query
      #if (None != self.contigs[ctg_id]): #{
      #  raise RealignerError("multiple alignments returned for contig: %s" %
      #    ctg_id)
      #} end if
      #new_align = MinimalAlignmentCls(align)
      #if (None == self.contigs[ctg_id] or
      #    self.contigs[ctg_id].match < new_align.match): #{
      #  if (None != self.contigs[ctg_id]): #{
      #    DebugMsg(self, "Replacing alignment")
      #  #} end if
      #  self.contigs[ctg_id] = new_align
      #} end if
      self.contigs[ctg_id].UpdateAligns(align)
    #} end for
  #} end def
#} end class

class RealignedContigCls: #{
  def __init__(self, ctg_id): #{
    self.id = ctg_id
    self.types = set()
    # region_pairs[member_id] = RegionPairCls
    self.region_pairs = dict()
    self.sequence = None
    self.best_align = None
    self.written = False
  #} end def

  def UpdateMembers(self, member, event_type): #{
    self.types.add(event_type)
    if ("fusion" == event_type): #{
      # update the region pairs
      self.region_pairs[member.IDString()] = RegionPairCls(member)
    #} end if
  #} end def

  def UpdateAligns(self, full_align): #{
    new_align = MinimalAlignmentCls(full_align)
    if (None == self.best_align or
        self.best_align.match < new_align.match): #{
      if (None != self.best_align): #{
        DebugMsg(self, "Replacing alignment (old: %i, new: %i)" %
          (self.best_align.match, new_align.match))
      #} end if
      self.best_align = new_align
    #} end if
    if (0 < len(self.region_pairs)): #{
      for region_pair in self.region_pairs.itervalues(): #{
        region_pair.UpdateGenes(full_align)
      #} end for
    #} end if
  #} end def
#} end class

class RegionPairCls: #{
  def __init__(self, member): #{
    self.member_id = member.IDString()
    self.starts = {'A': member.align_info_A.ctg_start,
      'B': member.align_info_B.ctg_start}
    self.ends   = {'A': member.align_info_A.ctg_end,
      'B': member.align_info_B.ctg_end}
    self.best_matches = {'A': 0, 'B': 0}
    self.gene_sets = {'A': set(), 'B': set()}
  #} end def

  def UpdateGenes(self, full_align): #{
    #align_span = (full_align.qend - full_align.qstart) + 1
    for region_id in ["A", "B"]: #{
      if (full_align.target in self.gene_sets[region_id]): #{
        continue
      #} end if
      #min_overlap = (min(self.Span(region_id), align_span) *
      #  MIN_REGION_OVERLAP_FRACTION)
      (overlap, fraction) = CalcOverlap(self.starts[region_id],
        self.ends[region_id], full_align.qstart, full_align.qend)
      match_diff = self.best_matches[region_id] - full_align.match
      if (MIN_REGION_OVERLAP_FRACTION < fraction and
          MAX_MATCH_DIFF > match_diff): #{
        self.gene_sets[region_id].add(full_align.target)
        self.best_matches[region_id] = max(self.best_matches[region_id],
          full_align.match)
      #} end if
    #} end for
  #} end def

  def GeneSetsOverlap(self): #{
    if (self.gene_sets['A'].isdisjoint(self.gene_sets['B'])): #{
      return False
    #} end if
    return True
  #} end def
#} end class

# to minimize the space required to store all the alignments, only
# store the necessary information
class MinimalAlignmentCls: #{
  def __init__(self, align): #{
    self.match = align.match
    self.ctg_start  = align.qstart
    self.ctg_end    = align.qend
    self.ctg_length = align.query_len
    self.num_query_gaps = align.qnuminsert
    # only save the blocks if there are query gaps that might need checking
    self.query_blocks = None
    if (0 < self.num_query_gaps): #{
      self.query_blocks = align.query_blocks
      # if aligned to the negative strand, need to reverse all the blocks
      if ("-" == align.query_strand): #{
        self.query_blocks.reverse()
        for index in range(len(self.query_blocks)): #{
          self.query_blocks[index].reverse()
        #} end for
      #} end if
    #} end if
    # if no more than 4 bases are not matches, mark a perfect alignment
    if (5 > (self.ctg_length - self.match)): #{
      self.perfect = True
    else:
      self.perfect = False
    #} end if
  #} end def

  def ToString(self): #{
    data_str = "%i-%i(%i)" % (self.ctg_start, self.ctg_end, self.match)
    if (0 < self.num_query_gaps): #{
      data_str += ": %s" % ",".join(["%i-%i" % (block[0],block[1]) for
        block in self.query_blocks])
    #} end if
    return data_str
  #} end if
#} end class

#### EXCEPTION CLASSES ####
class RealignerError(MyError): #{
  pass
#} end class
