#! /usr/bin/env python

# import standard modules
import os, re, string, sys, time, math
try: #{
  import ConfigParser
except ImportError:
  import configparser as ConfigParser
#} end try

# import custom modules
from log import OpenLogFile
from error import MyError
from messages import LogMsg, DebugMsg, ExtremeDebugMsg
from files_paths import EnsureAbsPath, CheckFilePath, PathError, FileBoxCls
from subprocesses import RunCommandFromString, STREAM_OUT

# constants
CFG_COMMANDS = "commands"

def SetupMainClass(self, options, params=None,
    log_file=None, log_info=None): #{
  # set general functions
  self.options = options
  self.debug = options.debug
  opts_string = "{%s}" % ", ".join([
    r"'%s': %s" % (field, getattr(options, field)) for
    field in sorted(vars(options).keys())])
  if (None != params): #{
    opts_string = str(params)
    self.params = params
  #} end if
  SetupLogInfo(self, options, opts_string, log_file, log_info)
  # ensure that the BARNACLE_PATH environmental variable is set
  SetupEnviron(self)
  SetupConfig(self)
  ExtremeDebugMsg(self, "Finished SetupMainClass")
#} end def

def SetupLogInfo(self, options, opts_string, log_file, log_info): #{
  self.close_log_file = False
  write_options = True
  if (None != log_info): #{
    if ('file' in log_info): #{
      log_file = log_info['file']
    #} end if
    if ('write_opts' in log_info): #{
      write_options = log_info['write_opts']
    #} end if
  #} end if
  # if no log file is given
  if (None == log_file): #{
    # attempt to open log file
    if (hasattr(options, 'log_file_name')): #{
      self.log_file = OpenLogFile(options.log_file_name)
      self.close_log_file = True
    else:
      self.log_file = None
    #} end if
    # write options being used to log file
    if (write_options): #{
      if (self.debug): #{
        LogMsg(self, opts_string)
      elif (None != self.log_file): #{
        self.log_file.WriteLine(opts_string)
      #} end if
    #} end if
  else:
    # use the given log file
    self.log_file = log_file
  #} end if
  self.log_info = {'debug': self.debug, 'file': self.log_file,
    'write_opts': False, 'extreme_debug': False}
  if (hasattr(options, 'extreme_debug')): #{
    self.log_info['extreme_debug'] = options.extreme_debug
    if (self.log_info['extreme_debug']): #{
      self.log_info['debug'] = True
    #} end if
  #} end if
#} end def

def SetupEnviron(self): #{
  if ("BARNACLE_PATH" in os.environ): #{
    return
  #} end if
  barnacle_dir = sys.path[0]
  DebugMsg(self, "Initially: %s" % barnacle_dir)
  if ("" == barnacle_dir): #{
    barnacle_dir = os.getcwd()
    DebugMsg(self, "Using current working dir: %s" % barnacle_dir)
  #} end if
  while (not os.path.isfile(os.path.join(barnacle_dir, "barnacle.pl"))): #{
    barnacle_dir = os.path.dirname(barnacle_dir)
    DebugMsg(self, "Up one dir: %s" % barnacle_dir)
    if ("" == barnacle_dir): #{
      raise MyError("cannot find BARNACLE directory")
    #} end if
  #} end while
  os.environ["BARNACLE_PATH"] = EnsureAbsPath(barnacle_dir)
  DebugMsg(self, "BARNACLE_PATH set to %s" % barnacle_dir)
#} end def

def SetupConfig(self): #{
  if (hasattr(self.options, "cfg")): #{
    return
  #} end if
  DebugMsg(self, "Loading BARNACLE config file...")
  config_path = os.path.join(os.environ["BARNACLE_PATH"], "barnacle.cfg")
  CheckFilePath(config_path, "BARNACLE configuration")
  cfg = ConfigParser.SafeConfigParser()
  cfg.readfp(open(config_path))
  if (not cfg.has_section(CFG_COMMANDS)): #{
    raise MyError("config file must contain %s section\nPath: %s" %
      (CFG_COMMANDS, config_path))
  #} end if
  cfg.path = config_path
  self.options.cfg = cfg
#} end def

def ConfigHasCommand(self, command): #{
  return self.options.cfg.has_option(CFG_COMMANDS, command)
#} end def

def CheckConfigCommands(self, required, local=True, check_path=True,
cluster_head=None): #{
  if (isinstance(required, str)): #{
    required = [required]
  #} end if
  missing_commands = list()
  for command in required: #{
    if (not local): #{
      if (None != cluster_head): #{
        cluster_command = command + "_%s" % cluster_head
        try: #{
          CheckConfigCommands(self, cluster_command,
            local=True, check_path=check_path)
          continue
        except (CommandMissingError, PathError), e:
          DebugMsg(self, "Checking general cluster command for %s" % command)
        #} end try
      #} end if
      cluster_command = command + "_cluster"
      try: #{
        CheckConfigCommands(self, cluster_command,
          local=True, check_path=check_path)
        continue
      except (CommandMissingError, PathError), e:
        DebugMsg(self, "Using local command for %s" % command)
      #} end try
    #} end if
    if (ConfigHasCommand(self, command)): #{
      if (check_path): #{
        path = GetCommand(self, command, local=True)
        CheckFilePath(path, "%s command" % command)
      #} end if
    else:
      missing_commands.append(command)
    #} end if
  #} end for
  if (0 < len(missing_commands)): #{
    raise CommandMissingError("The following required commands are missing "
      "from the config file: %s\nConfig path: %s" %
      (", ".join(missing_commands), self.options.cfg.path))
  #} end if
#} end def

def GetCommand(self, command, local=True, cluster_head=None): #{
  if (not local): #{
    if (None != cluster_head): #{
      cluster_command = command + "_%s" % cluster_head
      if (ConfigHasCommand(self, cluster_command)): #{
        return self.options.cfg.get(CFG_COMMANDS, cluster_command)
      #} end if
    #} end if
    cluster_command = command + "_cluster"
    if (ConfigHasCommand(self, cluster_command)): #{
      return self.options.cfg.get(CFG_COMMANDS, cluster_command)
    #} end if
  #} end if
  return self.options.cfg.get(CFG_COMMANDS, command)
#} end def

def GetGroupID(member_id): #{
  group_pattern = r"^(\d+)[a-z]+"
  group_match   = re.search(group_pattern, member_id)
  if (None == group_match): #{
    return None
  #} end if
  return int(group_match.group(1))
#} end def

def IntOrNAString(data): #{
  if ("N/A" == data): #{
    return data
  else:
    return "%i" % data
  #} end if
#} end def

def ConvertSeconds(seconds): #{
  days    = int(math.floor(seconds / 60.0 / 60.0 / 24.0))
  hours   = int(math.floor(seconds / 60.0 / 60.0 % 24.0))
  minutes = int(math.floor(seconds / 60.0 % 60.0))
  seconds = seconds % 60.0
  time_parts = list()
  if (0 < days): #{
    time_parts.append("%id" % days)
  #} end if
  if (0 < hours): #{
    time_parts.append("%ih" % hours)
  #} end if
  if (0 < minutes): #{
    time_parts.append("%im" % minutes)
  #} end if
  time_parts.append("%is" % seconds)
  return " ".join(time_parts)
#} end def

def IsNotUTR(exon_name): #{
  return not exon_name.lower().endswith(".utr")
#} end def

def IsNotNonCoding(exon_name): #{
  return not exon_name.lower().endswith(".non_coding")
#} end def

def IsNotUTRorNonCoding(exon_name): #{
  return (IsNotUTR(exon_name) and IsNotNonCoding(exon_name))
#} end def

def RemoveExonTypeLabel(exon_name): #{
  return exon_name.replace(".UTR","").replace(".non_coding","")
#} end def

def CleanExonName(exon_name): #{
  # remove the count
  exon_name = re.sub(r"X\d+X$", "", exon_name)
  # do NOT remove the part after the period, it is the exon type label
  # (e.g. UTR or non_coding)
  return exon_name.replace(" ", "_")
#} end def

def RunOverlapCode(target_coords_path, query_coords_path, output_path,
    dpt=False):
  # create the overlap command
  overlap_code = os.path.join(os.environ["BARNACLE_PATH"],
    "utils", "overlapcoordinates_ultrafast.pl")
  overlap_cmd = "%s -coord %s -ref %s" % (overlap_code,
    target_coords_path, query_coords_path)
  # open the output file
  fail_msg = "cannot open exon overlaps output file"
  if (STREAM_OUT == output_path): #{
    overlap_out_file = output_path
  else:
    overlap_out_file = {'path':output_path, 'mode':"w", 'fail_msg':fail_msg}
  #} end if
  # call the overlap command
  return RunCommandFromString(overlap_cmd, stdout=overlap_out_file,
    dpt=dpt)
#} end def

def TimeSpent(start_time): #{
  elapsed = time.time() - start_time
  hours   = int(math.floor(elapsed / 60 / 60))
  minutes = int(math.floor(elapsed / 60 % 60))
  seconds = elapsed % 60
  time_parts = list()
  if (0 < hours): #{
    time_parts.append("%ih" % hours)
  #} end if
  if (0 < minutes): #{
    time_parts.append("%im" % minutes)
  #} end if
  time_parts.append("%.3fs" % seconds)
  return "".join(time_parts)
#} end def

# return the reverse complement of a DNA sequence
def ReverseComplement(s): #{
  t = string.maketrans("AaCcGgTt", "TtGgCcAa")
  return s.translate(t)[::-1]
#} end def

class OptsCls: #{
  def __init__(self): #{
    self.debug = True
  #} end def
#} end class

def GetOptions(module): #{
  (options, arguments) = module.SetupOptionsParser().parse_args([""])
  return options
#} end def

def CheckOptions(options, required_member_list): #{
  missing_members = list()
  for member in required_member_list: #{
    if (not hasattr(options, member)): #{
      missing_members.append(member)
    #} end if
  #} end for
  if (0 < len(missing_members)): #{
    raise MyError("The following required members are missing:\n  %s" %
      "\n  ".join(missing_members))
  #} end if
#} end def

def IsEmpty(list_to_check): #{
  if (0 == len(list_to_check)): #{
    return True
  #} end if
  if (1 == len(list_to_check)): #{
    if ("n/a"  == list_to_check[0].lower() or
        "none" == list_to_check[0].lower()):
      return True
    #} end if
  #} end if
  return False
#} end def

class CommandMissingError(MyError): #{
  pass
#} end class

# check whether chromosome IDs in the given column of
# the given file use "chr"
def ShouldChromUseChr(column, path, file_description, log_info=None): #{
  fail_msg = "could not read %s coordinates file" % file_description
  file = FileBoxCls(path, "r", fail_msg)
  line = file.Head()
  chrom_value = line.split(" ")[column-1]
  if ("chr" in chrom_value): #{
    use_chr = True
  else:
    use_chr = False
  #} end if
  DebugMsg(log_info, "Line: \"%s\"\n" % line +
    "Chromosome Value: \"%s\" Use Chr: %s" % (chrom_value, use_chr))
  return use_chr
#} end def

def IsInteger(value): #{
  try: #{
    int_value = int(value)
  except ValueError:
    return False
  #} end try
  return True
#} end def

def WriteCommand(self, command_list): #{
  LogMsg(self, "COMMAND: %s" % " ".join(command_list), write_to_screen=False)
#} end def

def MaxMatchLength(query, target): #{
  max_match_len = 0
  # initialize last_row to a row of zeroes
  last_row = [0]*(len(target)+1)
  # initialize current_row to a single zero
  current_row = [0]
  #print "    %s" % " ".join(target)
  prev_char = " "
  for qend in range(len(query)): #{
    #print "%s %s" % (prev_char, " ".join(map(str,last_row)))
    prev_char = query[qend]
    for tend in range(len(target)): #{
      if (query[qend] == target[tend]): #{
        current_row.append(last_row[tend]+1)
        if (current_row[-1] > max_match_len): #{
          max_match_len = current_row[-1]
        #} end if
      else:
        current_row.append(0)
      #} end if
    #} end for
    last_row = current_row
    current_row = [0]
  #} end for
  #print "%s %s" % (query[qend], " ".join(map(str,last_row)))
  return max_match_len
#} end def

def AddChr(target): #{
  if (target.startswith("chr")): #{
    return target
  #} end if
  return "chr%s" % target
#} end if

def IsHomopolymerSequence(sequence): #{
  first = sequence[0]
  for character in sequence[1:]: #{
    if (character != first): #{
      return False
    #} end if
  #} end for
  return True
#} end def
